module Main where

import SigHeap
import HeapTree
import MPHeapTree
import RRHeapTree
import SkewHeapTree
import QuicksortST

import List(partition,sort)
import CPUTime

checkForce :: Ord a => [a] -> Bool
checkForce [] = True
checkForce [x] = True
checkForce (x:y:r) = x<=y && checkForce (y:r)



insertSort :: (Heap t, Ord a) => t a -> [a] -> [a]
insertSort nullTree xs = let heap = foldl (flip insert) nullTree xs
                         in fromHeap heap  

mergeSort :: Ord a => [a] -> [a]
mergeSort xs = mergeSort' (length xs) xs where
  mergeSort' n xs = 
    if n<2 then xs
           else let n' = n`div` 2
                    (ys,zs) = splitAt n' xs
                in merge (mergeSort' n' ys) (mergeSort' (n-n') zs)
       where merge (x:xs) (y:ys) = if x<=y then x : merge xs (y:ys)
                                           else y : merge (x:xs) ys
             merge [] ys = ys
             merge xs [] = xs
        


quickSortSimple :: Ord a => [a] -> [a]
quickSortSimple = qs where
   qs []     = []
   qs (x:xs) = let (ys,zs) = partition (<x) xs
               in qs ys ++ (x : qs zs)

quickSortFast :: Ord a => [a] -> [a]
quickSortFast xs = qs xs [] where
   qs []     acc = acc
   qs (x:xs) acc = let (ys,zs) = partition (<x) xs
                   in qs ys (x : qs zs acc)


singleTest :: Int -> ([Int] -> [Int]) -> IO String
singleTest n sf 
          = do
                let xs = [ i^2 `mod` (n`div`10) | i<-[0..n-1] ]::[Int]
                if (xs==xs) then do
                               start <- getCPUTime
                               let ys = sf xs
                               if checkForce ys 
                                  then do
                                        end <- getCPUTime
                                        let y = show ((end-start)`div`100000000000)
                                        return (init y ++"."++[last y])
                                  else return ("not sorted!")
                         else return ("?")
                             
               
seriesTest :: Algo -> [Int] -> IO ()
seriesTest (name,sf) ns  
   = do
      times <- mapM (\n -> singleTest n sf) ns 
      putStr (name ++ ": ") 
      mapM_ (\ (n,t) -> putStr ("("++show n++":"++t++"),")) (zip ns times)
      putStrLn ""

psort                    :: (Ord a) => [a] -> [a]
psort                    =  psortBy compare

psortBy                  :: (a -> a -> Ordering) -> [a] -> [a]
psortBy cmp              =  foldr (pinsertBy cmp) []

pinsert                  :: (Ord a) => a -> [a] -> [a]
pinsert                  = pinsertBy compare

pinsertBy                :: (a -> a -> Ordering) -> a -> [a] -> [a]
pinsertBy cmp x []       =  [x]
pinsertBy cmp x ys@(y:ys')
                        =  case cmp x y of
                                GT -> y : pinsertBy cmp x ys'
                                _  -> x : ys

type Algo = (String,[Int] -> [Int])

htree, mphtree, rrhtree, skewhtree, qsortSimple,qsortFast, ghcsort,msort,qsortst,psort1 :: Algo
htree = ("HTree",insertSort (empty::HTree Int)) 
mphtree = ("MPHTree", insertSort (empty::MPTree Int)) 
rrhtree = ("RRHTree", insertSort (empty::RRTree Int)) 
skewhtree = ("SkewHTree", insertSort (empty::SkewTree Int)) 
qsortSimple = ("QSortSlow", quickSortSimple) 
qsortFast = ("QSortFast", quickSortFast) 
ghcsort = ("GHCSort", List.sort) 
msort = ("mergesort", mergeSort) 
qsortst = ("quicksortST", quicksortST) 
psort1 = ("quicksortPrelude", psort) 

main :: IO ()
main = do
           let ser ns = mapM_ (flip seriesTest ns)
           ser [15000,30000,60000]   [qsortst,qsortFast]              -- 8/23
	   ser [5000,10000,20000]    [htree,mphtree,qsortFast]        -- 8/31
	   ser [50000,100000,200000] [qsortFast,mphtree,rrhtree,
                                      skewhtree,msort,ghcsort]        -- 8/41
           ser [1000*i | i<-[250,500,1000]] [skewhtree,msort,ghcsort] -- 8/42
 	  
{-
HTree: (3000:.6),(6000:2.9),(12000:17.0),
MPHTree: (3000:.0),(6000:.2),(12000:.5),
QSortSlow: (50000:3.4),(100000:11.6),(200000:42.8),
QSortFast: (50000:3.3),(100000:11.3),(200000:42.5),
MPHTree: (50000:3.0),(100000:8.5),(200000:24.0),
RRHTree: (50000:.8),(100000:2.3),(200000:6.5),
SkewHTree: (50000:.7),(100000:2.0),(200000:5.4),
mergesort: (50000:.6),(100000:1.4),(200000:3.4),
wGHCSort: (50000:.4),(100000:1.0),(200000:2.5),
-}
--          seriesTest skewhtree    [250000,500000,1000000]
--          seriesTest msort        [250000,500000,1000000]
--          seriesTest ghcsort      [250000,500000,1000000]
{-
SkewHTree: (250000:8.4),(500000:19.6),(1000000:51.7),
mergesort: (250000:4.4),(500000:10.2),(1000000:24.0),
GHCSort: (250000:3.7),(500000:8.8),(1000000:20.9),
-}      
{-
quicksortST: (3000:.1),(6000:.5),(12000:1.5),(25000:9.3),(50000:43.1),
-}
